
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "thingstack.h"

static void not_empty();

static thingstack_p stack=(void *)NULL;


static void not_empty(){
  if (stack==NULL){
    err_fatal("thingstack","attempt to pop empty stack");
  }
}

static thing_p new_thing(thing_type t){
  thing_p new;
  new=malloc(sizeof(thing_t));
  new->type=t;
  new->next=stack;
  stack=new;
  return new;
}

void push_string(char *str){
  thing_p thing;

  thing=new_thing(THING_STRING);
  thing->string=strdup(str);
}

void push_mark(){
  (void)new_thing(THING_MARK);
}

int top_is_mark(){
  return (stack->type==THING_MARK);
}
void push_integer(char *str){
  thing_p thing=new_thing(THING_INTEGER);

  thing->integer=strtol(str,NULL,10);

}

void push_int(int val){
  thing_p thing=new_thing(THING_INTEGER);

  thing->integer=val;

}
void push_rational(char *str){
  thing_p thing=new_thing(THING_RATIONAL);

  char *tail;
  thing->integer=strtol(str,&tail,10);
  while(!isdigit(*tail))
    tail++;

  thing->denominator=strtol(tail,NULL,10);
}

extern integer_t pop_int(){
  integer_t r;
  assert(stack && stack->type==THING_INTEGER);
  r=stack->integer;
  pop();
  return r;
}
    
extern void pop(){
  thing_p dead=stack;
  stack=stack->next;
  free(dead);
}
void stack_dump(){
  thing_p p=stack;
  while(p){
    switch(p->type){
    case THING_INTEGER:
      fputs("INTEGER\t",stdout);
      printf("%d\n",p->integer);
      break;
    case THING_RATIONAL:
      printf("RATIONAL\n %d/%d",p->integer,p->denominator);
      break;
    case THING_STRING:
      printf("STRING\t%s\n",p->string);
      break;
    }
    p=p->next;
  }

}
