
if (not IsBound(InfoPerturb)) then
    InfoPerturb:=NewInfoClass("Perturb");
    SetInfoLevel(InfoPerturb,1);
fi;

sortedOrbits:=function(G,S)
    return AsSet(List(Orbits(G,S),l->AsSet(l)));
end;



# Project the points V in ax=b
# This is not correct for arbitrary homogeneous coords.
project:=function(V,a,b)
    local p,M,dim,i,j,P,Vp;
    
    Vp:=StructuralCopy(V);
    
    dim:=Size(V[1])-1;
    
    M:=IdentityMat(dim+1);
    for i in [2..dim+1] do
        M[i][1]:=a[i];
        M[1][i]:=a[i];
    od;
    M[1][1]:=0;
    P:=1/M;
    for i in [1..Size(Vp)] do
        Vp[i][1]:=b;
    od;

    Vp:=List(Vp,r->P*r);
    for i in [1..Size(Vp)] do
        Vp[i][1]:=1;
    od;
    return Vp;
end;

# count the isometry orbits of (rows of) Mat 
countOrbits:=function(Mat,nauty)
    local gM,gg;
    gM:=InnerProdMat([1..Size(Mat)],Mat,DefaultInnerProduct);
    gg:=buildGramGraph([1..Size(Mat)],gM);
    gg:=canonGramViaNauty(gg,[1..Size(Mat)],1,nauty ) ;
    return Size(OrbitLengths(gg.automorphismGroup,[1..Size(Mat)]));
end;

groupSize:=function(Mat,nauty)
    local gg;
    gg:=groupViaGram3(Mat,DefaultInnerProduct,nauty ) ;
    return Order(gg);
end;

mostSymmetricProjection:=function(V,Candidates,nauty)
    local Vp,minOrbs,newOrbs,dir,bestDir;
    
    minOrbs:=Size(V);
    bestDir:=Candidates[1];
                 
    for dir in Candidates do
        Info(InfoPerturb,4,"trying ",dir);
        Vp:=project(V,dir,0);
        newOrbs:=countOrbits(Vp,nauty);
        Info(InfoPerturb,4,"got ", newOrbs, " orbits");
        if (newOrbs<minOrbs) then
            minOrbs:=newOrbs;
            bestDir:=dir;
        fi;
    od;
    return rec(dir:=bestDir,count:=minOrbs);
end;

mostSymmetricProjection2:=function(V,Candidates,nauty)
    local Vp,maxSize,newSize,dir,bestDir;
    
    maxSize:=groupSize(V,nauty);
    bestDir:=Candidates[1];
                 
    for dir in Candidates do
        Info(InfoPerturb,4,"trying ",dir);
        Vp:=project(V,dir,0);
        newSize:=groupSize(Vp,nauty);
        Info(InfoPerturb,4,"order= ", newSize);
        if (newSize>maxSize) then
            maxSize:=newSize;
            bestDir:=dir;
        fi;
    od;
    return rec(dir:=bestDir,size:=maxSize);
end;


orbitWiggle:=function(V,dir,nauty,innerproduct,epsilon)
    local i,o,Vp,Mat,gg,gM,orbits,orb,rho,N;
                 
    Mat:=project(V,dir,0);
    Vp:=StructuralCopy(V);
    
    Info(InfoPerturb,1,"Computing Automorphism Group: preprocessing");
    gM:=InnerProdMat([1..Size(Mat)],Mat,innerproduct);
    gg:=buildGramGraph([1..Size(Mat)],gM);
    Info(InfoPerturb,1,"Computing Automorphism Group: nauty");    
    gg:=canonGramViaNauty(gg,[1..Size(Mat)],1,nauty ) ;
    Info(InfoPerturb,1,"Computing Orbits");
    orbits:=Orbits(gg.automorphismGroup,[1..Size(Mat)]);
    
    N:=(1/epsilon);
    for orb in orbits do
        rho:=1+epsilon*Random([1..N])/N;
        for i in orb do
            Vp[i]:=Vp[i]*rho;
            Vp[i][1]:=1;
        od;
    od;
    return Vp;
end;


levelSets:=function(V,f)
    local vals,j,sets;
    
    vals:=AsSet(List(V,v->f(v)));
    sets:=[];
    for j in [1..Size(vals)] do
        sets[j]:=Filtered([1..Size(V)],i->f(V[i])=vals[j]);
    od;
    return sets;
end;

# find a subgroup of G that keeps together level sets (i.e. hyperplanes) of V normal to u 

projectGroup:=function(G,V,u,innerproduct)
    local vals,H,sizes,idx,S;
    vals:=AsSet(List(V,v->innerproduct(u,v)));
    
    H:=levelSets(V,v->innerproduct(u,v));
    Info(InfoPerturb,1,"Found ",Size(H), " level sets");
    sizes:=levelSets(H,r->Size(r));
    S:=G;
    for idx in sizes do
        Info(InfoPerturb,1,"Group size ",Order(S));
        S:=Stabilizer(S,H{idx},OnSetsSets);
    od;
   return S;
end;


weakProjectGroup:=function(G,V,u,innerproduct)
    local vals,H,flat,S;
    vals:=AsSet(List(V,v->innerproduct(u,v)));
    
    H:=levelSets(V,v->innerproduct(u,v));
    Info(InfoPerturb,1,"Found ",Size(H), " level sets");
    S:=G;
    for flat in H do
        Info(InfoPerturb,1,"Group size ",Order(S));
        S:=Stabilizer(S,flat,OnSets);
    od;
   return S;
end;


# mimic lex perturbation
orbitLex:=function(V,G,dir,innerproduct,epsilon )
    local i,Vp,Mat,S,gM,orbits,o,rho,pi,sigma;
    
    sigma:= -1;
    Mat:=project(V,dir,0);
    Vp:=StructuralCopy(V);
    
    Info(InfoPerturb,1,"Computing Automorphism Group: preprocessing");
    
    S:=projectGroup(G,V,dir,innerproduct);
    
    orbits:=sortedOrbits(S,[1..Size(Mat)]);
    Info(InfoPerturb,1,"Sorted orbits: ",orbits);
#    orbits:=Permuted(orbits,Random(SymmetricGroup(Size(orbits))));

    for o in [1..Size(orbits)] do
        rho:=1+sigma*epsilon^o;
        for i in orbits[o] do
            Vp[i]:=Vp[i]/rho;
            Vp[i][1]:=1;
        od;
    od;
    return rec(group:=S, matrix:=Vp);
end;

# mimic lex perturbation, but without recomputing 
# symmetry group

orbitLex2:=function(V,G,innerproduct,epsilon )
    local i,Vp,Mat,S,gM,orbits,o,rho,pi,sigma;
    
    sigma:= -1;
    Vp:=StructuralCopy(V);
    
    orbits:=sortedOrbits(G,[1..Size(Vp)]);

    Info(InfoPerturb,1,"Sorted orbits: ",orbits);
#    orbits:=Permuted(orbits,Random(SymmetricGroup(Size(orbits))));

    for o in [1..Size(orbits)] do
        rho:=1+sigma*epsilon^o;
        for i in orbits[o] do
            Vp[i]:=Vp[i]/rho;
            Vp[i][1]:=1;
        od;
    od;
    return rec(group:=G, matrix:=Vp);
end;

barrelize:=function(V,dir,inner,epsilon)
    local max,ip,v,i,Vb,rho;
    
    # solve a quadratic maximization problem :-)
    
    max:=0;
    
    dir[1]:=0;
    for v in V do
        ip:=inner(dir,v)^2;
        if ip>max then
            max:=ip;
        fi;
    od;
    
#    Print("max=",max);
    
    Vb:=[];
    for i in [1..Size(V)] do
        rho:=(1+(max-inner(dir,V[i])^2)*epsilon);
#        Print("rho=",rho,"\n");
        Vb[i]:=rho*V[i];
        Vb[i][1]:=1;
    od;
    return Vb;
end;
    
        
        
checkBasis:=function(basis,mat)
    local B,e,a;
    
    e:=List([1..Size(basis)],j->1);
    B:=List(mat{basis},r->r{[2..Size(basis)]});
    a:=(1/B)*e;
    return a;
end;

minOrbitsSubGroup:=function(G,S,m,tlim)
    
    local H,lastH,bestH,g,trial;
        
    bestH:=Group(());
    
    for trial in [1..tlim] do
        if (trial mod 100 =0) then
            Info(InfoDFS,2,"Trial ", trial);
        fi;
        
        H:=Group(());
        
        repeat
            g:=Random(G);
            lastH:=H;
            Info(InfoDFS,3,"adding generator ",g);        
            H:=ClosureGroup(H,g);
        until (Size(OrbitLengths(H,S))<m);

        if (Order(lastH) > Order(bestH)) then
            bestH:=lastH;
            Info(InfoDFS,2,"Found group of size ", Order(bestH));
        fi;
    od;
    return bestH;
end;

maxOrbitsSubGroup:=function(G,S,m,tlim)
    
    local H,lastH,bestH,g,trial,lastOrbs,orbs;
        
    bestH:=G;
        
    for trial in [1..tlim] do
        if (trial mod 100 =0) then
            Info(InfoDFS,2,"Trial ", trial);
        fi;
        H:=Group(());
        orbs:=Size(S);
        repeat
            g:=Random(G);
            lastH:=H;
            Info(InfoDFS,4,"adding generator ",g);        
            H:=ClosureGroup(H,g);
            lastOrbs:=orbs;
            orbs:=Size(OrbitLengths(H,S));
        until (orbs<m);
#        Info(InfoDFS,3,"orbs=", orbs, " lastOrbs=",lastOrbs);
        
        if (lastOrbs<=m and  Order(lastH) < Order(bestH)) then
            bestH:=lastH;
            Info(InfoDFS,2,"Found group of size ", Order(bestH));
        fi;
    od;
    return bestH;
end;

    
    
oppositeIndex:=function(V,i)
	local j,u;

	for j in [1..Size(V)] do
	  	u:= V[j]+V[i];
		u[1]:=0;
		if u*u=0 then
			return j;
		fi;	
	od;
	return fail;
end;
	

bestLineStabilizer:=function(G,V)
	local bestH,i,j,H;
	bestH:=Group(());
	for i in [1..Size(V)] do	
		j:=oppositeIndex(V,i);
		if j<>fail then
			H:=Stabilizer(G,[i,j],OnSets);
			if (Order(H) > Order(bestH)) then	
				Info(InfoDFS,2,"line ",i," Order ",Order(H),"\n");
				bestH:=H;
			fi;
		fi;
	od;
	return bestH;
end;

